# -*- coding: utf-8 -*-
import getpass
import json
import logging
import os
import socket
import sys
from datetime import datetime
from pathlib import Path

import cv2
import numpy as np
from arkindex import ArkindexClient
from line_image_extractor.image_utils import BoundingBox

import yaml

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s %(levelname)s/%(name)s: %(message)s"
)
logger = logging.getLogger(os.path.basename(__file__))


if not os.environ.get("PYTHONHASHSEED"):
    logger.info("Starting again with PYTHONHASHSEED for reproducibility")
    os.environ["PYTHONHASHSEED"] = "0"
    os.execv(sys.executable, ["python3"] + sys.argv)


class TranscriptionData:
    def __init__(
        self,
        element_id,
        element_name,
        polygon,
        text,
        trans_id,
        worker_version_id,
        rotation_class=None,
    ):
        self.element_id = element_id
        self.element_name = element_name
        self.polygon = np.asarray(polygon).clip(0)
        self.text = text
        self.trans_id = trans_id
        self.worker_version_id = worker_version_id
        self.rotation_class = rotation_class

        self.rect = BoundingBox._make(cv2.boundingRect(self.polygon))

    @property
    def is_vertical(self) -> bool:
        """
        Used to filter out vertical lines. Will be ignored when rotation class is given.
        """
        if self.rotation_class is None:
            return self.rect.height > self.rect.width
        return False

    def __repr__(self):
        return str(vars(self))

    @classmethod
    def copy_replace_polygon(cls, trans: "TranscriptionData", new_polygon):
        """
        Class method to keep the change logic inside the class - less likely to forget to update.
        """
        return TranscriptionData(
            trans.element_id,
            trans.element_name,
            new_polygon,
            trans.text,
            trans.trans_id,
            trans.worker_version_id,
            trans.rotation_class,
        )


def write_file(file_name, content):
    with open(file_name, "w") as f:
        f.write(content)


class CachedApiClient(ArkindexClient):
    def __init__(self, cache_root: Path = None, **kwargs):
        logger.info("Creating cached api client")
        super().__init__(**kwargs)
        self.cache_root = cache_root

        self.server_name = self.document.url.split("//")[-1].split(".")[0]
        logger.info(f"Server is {self.server_name}")

        self.cache_location = self.cache_root / "requests" / self.server_name

        self.cache_location.mkdir(parents=True, exist_ok=True)

    def cached_paginate(self, *args, **kwargs):
        logger.info(f"Params: args: {args} ---- kwargs: {kwargs}")

        # expecting only one positional argument - operation_id
        if len(args) == 1:
            operation_id = args[0]
            logger.info(f"op: {operation_id}")
        else:
            raise ValueError(f"Unexpected number of positional arguments: {args}")

        if "id" in kwargs:
            ark_id = kwargs["id"]
        elif "corpus" in kwargs:
            ark_id = kwargs["corpus"]
        else:
            raise ValueError(f"Id or corpus must be defined: {kwargs}")

        kwargs_hash = hash(json.dumps(kwargs, sort_keys=True))
        logger.info(f"Kwargs hash: {kwargs_hash}")

        request_cache = (
            self.cache_location / operation_id / ark_id / f"hash_{kwargs_hash}.json"
        )
        if request_cache.exists():
            logger.info(f"Loading from cache: {operation_id} - {ark_id}")
            cached_results = json.loads(request_cache.read_text())
            for res in cached_results:
                yield res
            logger.info("Used cached results")
            return

        logger.info(f"Running actual query: {operation_id} - {ark_id}")
        results = []
        paginator = self.paginate(*args, **kwargs)
        for res in paginator:
            # logger.info(f"Res: {res}")
            results.append(res)
            yield res

        logger.info(f"Saving to cache: {operation_id} - {ark_id}")
        request_cache.parent.mkdir(parents=True, exist_ok=True)
        request_cache.write_text(json.dumps(results))
        logger.info("Saved")


def write_json(d, filename):
    with open(filename, "w") as f:
        f.write(json.dumps(d, indent=4))


def write_yaml(d, filename):
    with open(filename, "w") as f:
        yaml.dump(d, f)


def export_parameters(
    format,
    dataset_name,
    out_dir,
    common,
    image,
    split,
    select,
    filter,
    datasets,
    arkindex_api_url,
):
    """
    Dump a JSON log file to keep track of parameters for this dataset
    """
    # Get config dict
    config = {
        "format": format,
        "dataset_name": dataset_name,
        "out_dir": out_dir,
        "common": vars(common),
        "image": vars(image),
        "split": vars(split),
        "select": vars(select),
        "filter": vars(filter),
    }
    # Serialize special classes
    image.extraction_mode = image.extraction_mode.name
    filter.transcription_type = filter.transcription_type.value
    common.cache_dir = str(common.cache_dir)

    if common.log_parameters:
        # Get additional info on dataset and user
        current_datetime = datetime.now().strftime("%Y-%m-%d|%H-%M-%S")

        parameters = {}
        parameters["config"] = config
        parameters["dataset"] = (
            datasets
            if isinstance(datasets, dict)
            else {"train": datasets[0], "valid": datasets[1], "test": datasets[2]}
        )
        parameters["info"] = {
            "user": getpass.getuser(),
            "date": current_datetime,
            "device": socket.gethostname(),
            "arkindex_api_url": arkindex_api_url,
        }

        # Export
        parameter_file = os.path.join(
            out_dir, f"param-{dataset_name}-{format}-{current_datetime}.json"
        )
        write_json(parameters, parameter_file)
        logger.info(f"Parameters exported in file {parameter_file}")

    config_file = os.path.join(out_dir, "config.yaml")
    write_yaml(config, config_file)
    logger.info(f"Config exported in file {config_file}")
