# -*- coding: utf-8 -*-
import random
from pathlib import Path

import pytest

from kaldi_data_generator.arguments import FilterArgs
from kaldi_data_generator.main import MANUAL, HTRDataGenerator, KaldiPartitionSplitter


def test_init():
    htr_data_gen = HTRDataGenerator(format="kaldi")
    assert htr_data_gen is not None


@pytest.mark.parametrize(
    "worker_version_ids, expected_trans_lines",
    (
        ([], 55),
        ([MANUAL], 55),  # only manual transcriptions in this example
        (["test_1234"], 0),  # no transcriptions with this worker version
        (["test_1234", MANUAL], 55),  # searching by worker version in this order
    ),
)
def test_run_volumes_with_worker_version(
    fake_run_volume_api_client,
    fake_volume_id,
    fake_image,
    worker_version_ids,
    expected_trans_lines,
    tmpdir,
    mocker,
):
    api_client = fake_run_volume_api_client
    out_dir_base = tmpdir
    htr_data_gen = HTRDataGenerator(
        format="kaldi",
        filter=FilterArgs(accepted_worker_version_ids=worker_version_ids),
        out_dir_base=out_dir_base,
    )
    htr_data_gen.api_client = api_client

    htr_data_gen.get_image = mocker.MagicMock()
    # return same fake image for all the pages
    htr_data_gen.get_image.return_value = fake_image
    htr_data_gen.run_volumes([fake_volume_id])

    trans_files = list(Path(htr_data_gen.out_line_text_dir).glob("*.txt"))
    # assert files aren't empty
    assert all(len(trans_file.read_text().strip()) > 0 for trans_file in trans_files)
    assert len(trans_files) == expected_trans_lines

    # each image file should have one transcription file
    img_files = list(Path(htr_data_gen.out_line_img_dir).glob("*.jpg"))
    assert len(img_files) == len(trans_files)

    assert len(api_client.history) == 9
    assert api_client.responses == []


def test_get_accepted_zones_filter_metadata(tmpdir, fake_run_filter_metadata):
    api_client = fake_run_filter_metadata
    out_dir_base = tmpdir
    htr_data_gen = HTRDataGenerator(
        format="kaldi",
        filter=FilterArgs(accepted_metadatas={"Language": "arabic"}),
        out_dir_base=out_dir_base,
    )
    htr_data_gen.api_client = api_client

    assert htr_data_gen.get_accepted_zones("fake_page") == [
        "0b719a5a-40c7-47ec-96c8-6c3064df5485",
        "18725942-24f5-4f81-a16f-62c1323c1041",
        "3ebe7e48-fe2f-4533-92df-9895db05c3f5",
    ]


def test_create_partitions(fake_expected_partitions, tmpdir):
    out_dir_base = Path(tmpdir)
    splitter = KaldiPartitionSplitter(out_dir_base=out_dir_base)

    all_ids = []
    for split_name, expected_split_ids in fake_expected_partitions.items():
        all_ids += expected_split_ids

    # shuffle to show that splits are reproducible
    random.shuffle(all_ids)

    # just to check that there's no problem with the data
    assert len(all_ids) > 0

    partitions_dir = splitter.out_dir_base / "Partitions"
    lines_dir = splitter.out_dir_base / "Lines"
    trans_dir = splitter.out_dir_base / "Transcriptions"

    # create fake data
    for line_id in all_ids:
        trans_file = trans_dir / f"{line_id}.txt"
        trans_file.parent.mkdir(parents=True, exist_ok=True)
        trans_file.touch()

        img_file = lines_dir / f"{line_id}.jpg"
        img_file.parent.mkdir(parents=True, exist_ok=True)
        img_file.touch()

    splitter.create_partitions()

    for split_name, expected_split_ids in fake_expected_partitions.items():
        part_file = partitions_dir / split_name
        assert part_file.read_text().strip().split("\n") == expected_split_ids
